{% extends "sphinx_rtd_theme/layout.html" %}
{% set css_files = css_files + ['_static/basic.css'] %}

{%- block sidebartitle %}

{%- if logo and theme_logo_only %}
<a href="{% if READTHEDOCS %}{{ pathto(master_doc) }}{% else %}/#{% endif %}">
{%- else %}
<a href="{% if READTHEDOCS %}{{ pathto(master_doc) }}{% else %}/#{% endif %}" class="icon icon-home"> {{ project }}
{%- endif %}

{%- if logo %}
{#- Not strictly valid HTML, but it's the only way to display/scale
    it properly, without weird scripting or heaps of work
#}
<img src="{{ logo_url }}" class="logo" alt="{{ _('Logo') }}"/>
{%- endif %}
</a>

{%- if theme_display_version %}
{%- set nav_version = version %}
{%- if READTHEDOCS and current_version %}
  {%- set nav_version = current_version %}
{%- endif %}
{%- if nav_version %}
  <div class="version">
    {{ nav_version }}
  </div>
{%- endif %}
{%- endif %}

{%- include "searchbox.html" %}

{%- endblock %}
